#include <s3e.h>
#include <s3eSocket.h>
#include <s3eTimer.h>

#include <vector>
#include <list>
#include <set>
#include <map>
#include <string>

#include <airplay-multi-resolver.h>

/*
 * The Airplay async resolver is not concurrent, so we should
 * implement resolver requests queue
*/
class _request_queue_entry {
	std::set< std::pair<airplay_multires_callback,void* > > callbacks;
	std::string request;
public:
	_request_queue_entry() {}
	_request_queue_entry(const std::string &_request)
		: request(_request)
	{}
	_request_queue_entry(const std::string &_request, airplay_multires_callback callback,void *userdata )
		: request(_request)
	{
		add_callback(callback,userdata);
	}
	_request_queue_entry(const _request_queue_entry &entry)
		: request(entry.request),callbacks(entry.callbacks)
	{}
	virtual ~_request_queue_entry()
	{}
	void add_callback(airplay_multires_callback callback,void *userdata)
	{
		callbacks.insert(std::pair<airplay_multires_callback,void*>(callback,userdata));
	}
	bool remove_callback(airplay_multires_callback callback,void *userdata)
	{
		std::set< std::pair<airplay_multires_callback,void* > >::const_iterator e = callbacks.end();
		std::set< std::pair<airplay_multires_callback,void* > >::const_iterator i = callbacks.find(std::pair<airplay_multires_callback,void*>(callback,userdata));
		if( i != e ) {
			callbacks.erase(i);
			return true;
		}
		return false;
	}
	bool callback_present(airplay_multires_callback callback,void *userdata) const
	{
		std::set< std::pair<airplay_multires_callback,void* > >::const_iterator e = callbacks.end();
		std::set< std::pair<airplay_multires_callback,void* > >::const_iterator i = callbacks.find(std::pair<airplay_multires_callback,void*>(callback,userdata));
		if( i != e ) {
			return true;
		}
		return false;
	}
	void set_request(const std::string &s)
	{
		request = s;
	}
	std::string get_request() const
	{
		return request;
	}
	unsigned callback_count() const
	{
		return callbacks.size();
	}
	void success( s3eInetIPAddress resolved )
	{
		std::set< std::pair<airplay_multires_callback,void* > > copy = callbacks; // for safe iteration
		std::set< std::pair<airplay_multires_callback,void* > >::const_iterator e = copy.end();
		std::set< std::pair<airplay_multires_callback,void* > >::const_iterator i = copy.begin();
		for( ; i != e; i++ ) {
			i->first(i->second,resolved);
		}
	}
	void failure()
	{
		std::set< std::pair<airplay_multires_callback,void* > > copy = callbacks; // for safe iteration
		std::set< std::pair<airplay_multires_callback,void* > >::const_iterator e = copy.end();
		std::set< std::pair<airplay_multires_callback,void* > >::const_iterator i = copy.begin();
		for( ; i != e; i++ ) {
			i->first(i->second,0);
		}
	}
};

class _request_queue {
	std::vector< _request_queue_entry > queue;
	std::vector< std::pair<_request_queue_entry,s3eInetIPAddress> > results;
	s3eInetAddress buffer;
	bool lookup_started;
	_request_queue() : lookup_started(false) {}

	void success(s3eInetAddress *addr) {
		std::vector< _request_queue_entry >::iterator e = queue.end();
		std::vector< _request_queue_entry >::iterator b = queue.begin();
		if( b != e ) {
			results.push_back(std::pair<_request_queue_entry,s3eInetIPAddress>(*b,addr->m_IPAddress));
			queue.erase(b);
		}
	}

	void failure() {
		std::vector< _request_queue_entry >::iterator e = queue.end();
		std::vector< _request_queue_entry >::iterator b = queue.begin();
		if( b != e ) {
			results.push_back(std::pair<_request_queue_entry,s3eInetIPAddress>(*b,0));
			queue.erase(b);
		}
	}

	int _release_results() {
		std::vector< std::pair<_request_queue_entry,s3eInetIPAddress> > copy = results; // for safe iterations
		std::vector< std::pair<_request_queue_entry,s3eInetIPAddress> >::iterator e = copy.end();
		std::vector< std::pair<_request_queue_entry,s3eInetIPAddress> >::iterator i = copy.begin();
		for( ; i != e; i++ ) {
			if( i->second )
				i->first.success(i->second);
			else
				i->first.failure();
		}
		int size = (int)copy.size();
		results.clear();
		return size;
	}

	static int timer_callback(void *sys, void *user)
	{
		return lookup_callback(user,&instance());
	}

	static int lookup_callback(void *sys, void *user)
	{
		if( sys ) {
			instance().success( (s3eInetAddress *)sys );
		} else {
			instance().failure();
		}
		instance().lookup_started = false;
		instance().ensure_lookups();
		return 0;
	}

	void start_lookups()
	{
		std::vector< _request_queue_entry >::const_iterator e = queue.end();
		std::vector< _request_queue_entry >::const_iterator b = queue.begin();
		if( b != e ) {
			lookup_started = true;
			s3eInetLookup( b->get_request().c_str(),&buffer,lookup_callback,this);
			//s3eInetLookup( b->get_request().c_str(),&buffer,NULL,NULL);
			//s3eTimerSetTimer(200,timer_callback,&buffer);
		}
	}
	void ensure_lookups()
	{
		if( !lookup_started )
			start_lookups();
	}
	void _enqueue(const std::string &host,airplay_multires_callback cb, void *userdata)
	{
		std::vector< _request_queue_entry >::iterator e = queue.end();
		std::vector< _request_queue_entry >::iterator b = queue.begin();
		std::vector< _request_queue_entry >::iterator i = b;
		for( ; i != e; i++ ) {
			if( i->get_request() == host ) {
				i->add_callback(cb,userdata);
				return;
			}
		}
		queue.push_back(_request_queue_entry(host,cb,userdata));
		ensure_lookups();
	}
	bool _unqueue(const std::string &host,airplay_multires_callback cb, void *userdata)
	{
		std::vector< _request_queue_entry >::iterator e = queue.end();
		std::vector< _request_queue_entry >::iterator b = queue.begin();
		std::vector< _request_queue_entry >::iterator i = b;
		for( ; i != e; i++ ) {
			if( i->get_request() == host ) {
				bool r = i->remove_callback(cb,userdata);
				bool start_again = i == b && !i->callback_count();
				if( !i->callback_count() ) {
					queue.erase(i);
				}
				if( start_again ) {
					s3eInetLookupCancel();
					ensure_lookups();
				}
				return r;
			}
		}
		return false;
	}
	bool _is_queued(const std::string &host,airplay_multires_callback cb, void *userdata)
	{
		std::vector< _request_queue_entry >::const_iterator e = queue.end();
		std::vector< _request_queue_entry >::const_iterator b = queue.begin();
		std::vector< _request_queue_entry >::const_iterator i = b;
		for( ; i != e; i++ ) {
			if( i->get_request() == host ) {
				return i->callback_present(cb,userdata);
			}
		}
		return false;
	}

	static _request_queue *_instance;
public:
	static _request_queue &instance()
	{
		initialize();
		return *_instance;
	}
	static void initialize()
	{
		if( !_instance )
			_instance = new _request_queue;
	}
	static void cleanup()
	{
		if( _instance ) {
			if( !_instance->queue.empty() ) {
				s3eInetLookupCancel();
			}
			delete _instance;
			_instance = 0;
		}
	}
	static void enqueue(const std::string &host,airplay_multires_callback cb, void *userdata)
	{
		_request_queue::instance()._enqueue(host,cb,userdata);
	}
	static bool unqueue(const std::string &host,airplay_multires_callback cb, void *userdata)
	{
		return _request_queue::instance()._unqueue(host,cb,userdata);
	}
	static bool is_queued(const std::string &host,airplay_multires_callback cb, void *userdata)
	{
		return _request_queue::instance()._is_queued(host,cb,userdata);
	}
	static int release_results()
	{
		return instance()._release_results();
	}
};

_request_queue *_request_queue::_instance;

void airplay_multires_enqueue(const char *hostname,airplay_multires_callback cb,void *userdata)
{
	_request_queue::enqueue(hostname,cb,userdata);
}
int airplay_multires_unqueue(const char *hostname,airplay_multires_callback cb,void *userdata)
{
	return _request_queue::unqueue(hostname,cb,userdata);
}
int airplay_multires_is_queued(const char *hostname,airplay_multires_callback cb,void *userdata)
{
	return _request_queue::is_queued(hostname,cb,userdata);
}
int airplay_multires_release_results()
{
	return _request_queue::release_results();
}

void airplay_multires_init()
{
	_request_queue::initialize();
}
void airplay_multires_cleanup()
{
	_request_queue::cleanup();
}
